part of snowm_scanner;

class _SnowMScanner {
  static const MethodChannel _channel = const MethodChannel('snowm_scanner');
  int _beaconScannerCount = 0;
  int _stateScannerCount = 0;
  Map<String, StreamController<List<SnowMBeacon>>> _beaconStreamPool = {};
  Map<String, StreamController<BluetoothState>> _stateStreamPool = {};
  Map<String, Map<String, SnowMBeacon>> _cachedBeacons = {};

  bool _enableMqtt = false;
  int _timeInterval = 600000;

  _SnowMScanner() {
    _channel.setMethodCallHandler((call) {
      String method = call.method;
      if (method.contains("scanBeacons") &&
          _beaconStreamPool.containsKey(call.method)) {
        List<SnowMBeacon> beacons = call.arguments["beacons"]
            .map<SnowMBeacon>((b) => SnowMBeacon._fromMap(b))
            .toList();
        _beaconStreamPool[call.method].add(beacons);
      } else if (method.contains("bluetoothStateListener")) {
        _stateStreamPool[call.method]
            .add(getBluetoothStateFromKey(call.arguments));
      }
      return;
    });
  }

  void configure({
    /// Should scanner send messages to mqtt protocol. Default [false]
    bool enableMqtt = false,

    /// Time interval[Milliseconds] to send data to IOT cloud.Default is 10 minutes
    /// Valid only if [enableMqtt] is true.
    int timeInterval = 600000,
  }) {
    assert(timeInterval >= 60000);
    this._enableMqtt = enableMqtt;
    this._timeInterval = timeInterval;
  }

  Stream<BluetoothState> getBluetoothStateStream() {
    String method = "bluetoothStateListener#${++_stateScannerCount}";
    _stateStreamPool[method] =
        _getNewBluetoothStateStreamController<BluetoothState>(method);
    _channel.invokeMethod(method);
    return _stateStreamPool[method].stream;
  }

  Future<BluetoothPermissionState> getPermissionState() async {
    return getBluetoothPermissionStateFromKey(
        await _channel.invokeMethod<String>("permissionState"));
  }

  // TODO: Make it asyncronous
  void requestPermission() {
    _channel.invokeMethod<String>("requestPermission");
  }

  Future<BluetoothState> getBluetoothState() async {
    return getBluetoothStateFromKey(
        await _channel.invokeMethod<String>("bluetoothState"));
  }

  Future<void> scanBeaconsInBackground(
    List<String> uuids, {

    /// [Android] Title of the foreground scanner notification title
    String title = "SnowM Scanner",

    /// [Android] Title of the foreground scanner notification body
    String body = "We are scanning beacons currently.",

    /// Custom payload that can be sent to scanner which will be delivered along with the scanned beacons
    Map<String, dynamic> customData = const {},
  }) async {
    assert(uuids.length > 0);
    await _channel.invokeMethod('scanBeaconBackground', {
      "uuids": uuids.map((e) => e.toUpperCase()).toList(),
      "customData": customData,
      "title": title,
      "body": body,
    });
  }

  Future<void> stopBackgroudScan() async {
    await _channel.invokeMethod('stopBackgroundScan');
  }

  Stream<List<SnowMBeacon>> scanBeacons({
    /// List of proximity uuid of a beacon to be scanned
    List<String> uuids,

    /// Milliseconds for which the beacons should be cached. Default value is 60 seconds.
    int cacheBeaconsFor = 10000,
  }) {
    String method = "scanBeacons#${++_beaconScannerCount}";
    _beaconStreamPool[method] =
        _getNewBeaconStreamController<List<SnowMBeacon>>(method);
    _channel.invokeMethod(method, {
      "uuids": uuids.map((e) => e.toUpperCase()).toList(),
      "enableMqtt": _enableMqtt,
      "timeInterval": _timeInterval
    });
    Map<String, SnowMBeacon> cachedBeacons = _cachedBeacons[method] ?? {};
    return _beaconStreamPool[method].stream.map((beacons) {
      int currentTime = DateTime.now().millisecondsSinceEpoch;
      beacons = beacons.map((e) {
        e.._detectedTime = currentTime;
        cachedBeacons[e.uuid] = e;
        return e;
      }).toList();
      List<SnowMBeacon> freshBeacons = cachedBeacons.values
          .where((b) => (currentTime - b._detectedTime) < cacheBeaconsFor)
          .toList();
      cachedBeacons.clear();
      freshBeacons.forEach((b) {
        cachedBeacons[b.uuid] = b;
      });
      return freshBeacons;
    });
  }

  StreamController<T> _getNewBeaconStreamController<T>(String identifier) {
    StreamController<T> controller = StreamController<T>();
    controller.onCancel = () {
      _channel.invokeMethod("cancelStream", {
        "methodName": identifier,
      });
      _beaconStreamPool.remove(identifier);
    };
    return controller;
  }

  StreamController<T> _getNewBluetoothStateStreamController<T>(
      String identifier) {
    StreamController<T> controller = StreamController<T>();
    controller.onCancel = () {
      _channel.invokeMethod("cancelBluetoothStateListener", {
        "methodName": identifier,
      });
      _beaconStreamPool.remove(identifier);
    };
    return controller;
  }
}

final snowmScanner = _SnowMScanner();
